#include "lib_ll.h"

void check_args(int argc, char** argv) {
    
    if (argc != 5) {
	fprintf(stderr,"USAGE:  %s <L> <rho> <seed/nstats> <savefile>\n",argv[0]);
	exit(-1);
    }
    
    return;
    
}

void create_lattice(struct lattice* R, int L, double rho, int rnum) {


    int i,V,x,y;
    unsigned long long rn;

    R->rho=rho;
    rn=rnum;

    R->L=L;
    V=L*L;

    if (R->rho > 1) {
	fprintf(stderr,"Wrong density spec: %lf\n",R->rho);
	exit(-2);
    }

    if (NULL==(R->site=(struct cluster*)calloc(V,sizeof(struct cluster)))) {
	fprintf(stderr,"Can't alloc lattice memory\n");
	exit(-2);
    }

    if (NULL==(R->oXp=(int*)calloc(L,sizeof(int))) ||
	NULL==(R->oXm=(int*)calloc(L,sizeof(int))) ||
	NULL==(R->oYp=(int*)calloc(L,sizeof(int))) ||
	NULL==(R->oYm=(int*)calloc(L,sizeof(int))) ) {
	fprintf(stderr,"Can't alloc lattice memory\n");
	exit(-2);
    }

    for (i=0;i<L-1;i++) {
	R->oXp[i]= 1;
	R->oYp[i]= L;
	R->oXm[L-i-1]= -1;
	R->oYm[L-i-1]= -L;
    }

    R->oXp[L-1]=-(L-1);
    R->oYp[L-1]=-L*(L-1);
    R->oXm[0]  = L-1;
    R->oYm[0]  = L*(L-1);

    for (i=0;i<10000;i++) rn=ECUYER(rn);    
    for (y=0,i=0;y<L;y++) {
	for (x=0;x<L;x++,i++) {
	    R->site[i].value=( NORM(rn) < (R->rho) ? 0 : 1 );
	    R->site[i].next =(R->site)+i;
	    R->site[i].prev =(R->site)+i;
	    R->site[i].first=(R->site)+i;
	    R->site[i].last =(R->site)+i;
	    R->site[i].mass =1;
	    R->site[i].xi2  =x*x+y*y;
	    R->site[i].x    =x;
	    R->site[i].y    =y;
	    R->site[i].cmx  =x;
	    R->site[i].cmy  =y;
	}
    }
#ifdef PRINTOUT
    print_values(R);
#endif

    return ;

}

void destroy_lattice(struct lattice* R) {

    if (NULL!=R->site) free(R->site);
    if (NULL!=R->oXp)   free(R->oXp);
    if (NULL!=R->oXm)   free(R->oXm);
    if (NULL!=R->oYp)   free(R->oYp);
    if (NULL!=R->oYm)   free(R->oYm);

    return;

}


void print_values(struct lattice* R) {

    int x,y,i;
    fprintf(stdout, "#######");
    fprintf(stdout,"\n");
    for(y=0,i=0 ; y<R->L ; y++) {
	for(x=0;x<R->L; x++, i++) {
	    fprintf(stdout, "%5d  ",R->site[i].value);
	}
	fprintf(stdout,"\n");
    }
    fprintf(stdout, "#######");
    fprintf(stdout,"\n");

}

void print_labels(struct lattice* R) {

    int x,y,i;
    fprintf(stdout, "#######");
    fprintf(stdout,"\n");
    for(y=0,i=0 ; y<R->L ; y++) {
	for(x=0;x<R->L; x++, i++) {
	    fprintf(stdout, "%5X  ",(unsigned int)(R->site[i].first-R->site));
	}
	fprintf(stdout,"\n");
    }
    fprintf(stdout, "#######");
    fprintf(stdout,"\n");

}

void save_lattice(struct lattice* R, char* savefile){

    int x,y,i;
    FILE* fout;

    fout=fopen(savefile,"w");

    fprintf(fout, "## x y label value\n");

    for(y=0,i=0 ; y<R->L ; y++) {
	for(x=0;x<R->L; x++, i++) {
	    fprintf(fout, "%10d %10d %10d %10d\n",x,y,(int)(R->site[i].first-R->site),R->site[i].value);
	}
    }

    fclose(fout);

    return;

}

int propagate_labels(struct lattice* R) {

    int x,y,i,howmany;
    int yplus, yminus;
    int xplus, xminus;

    for(y=0,i=0,howmany=0 ; y<R->L ; y++) {
	yplus=y*R->L+R->oYp[y];
	yminus=y*R->L+R->oYm[y];
	for(x=0;x<R->L; x++, i++, yplus++, yminus++) {
	    xplus=i+R->oXp[x];
	    xminus=i+R->oXm[x];
#ifndef PBC
	    if (x<R->L-1) howmany+=update(R,i,xplus);
	    if (x>0)   howmany+=update(R,i,xminus);
	    if (y<R->L-1) howmany+=update(R,i,yplus);
	    if (y>0)   howmany+=update(R,i,yminus);
#endif
#ifdef PBC
	    howmany+=update(R,i,xplus);
	    howmany+=update(R,i,xminus);
	    howmany+=update(R,i,yplus);
	    howmany+=update(R,i,yminus);
#endif
	}
    }

    return howmany;
}


int update(struct lattice* R, int here, int neighbor) {

    int winner=-1, loser=-1;
    struct cluster* dummy;

    struct cluster* loserFirst;
    struct cluster* loserLast;
    struct cluster* winnerFirst;
    struct cluster* winnerLast;

    struct cluster* runner;

    dummy=R->site;

    if (dummy[here].value == dummy[neighbor].value ) {

	if (dummy[here].first == dummy[neighbor].first) {
	    return 0;
	} else if (dummy[here].first > dummy[neighbor].first) {
	    winner=neighbor; 
	    loser=here;
	} else if (dummy[here].first < dummy[neighbor].first) {
	    winner=here; 
	    loser=neighbor;	    
	}

	winnerFirst=dummy[winner].first;
	winnerLast =dummy[winner].last;
	loserFirst=dummy[loser].first;
	loserLast =dummy[loser].last;

	runner=loserFirst;
	runner->first=winnerFirst;

	while (runner != runner->next) {
	    runner=runner->next;
	    runner->first=winnerFirst;
	}

	runner=winnerFirst;
	runner->last=loserLast;
	while (runner != runner->next) {
	    runner=runner->next;
	    runner->last=loserLast;
	}

	winnerLast->next=loserFirst;
	loserFirst->prev=winnerLast;
	
	winnerFirst->mass+=loserFirst->mass;
#ifndef PBC
	winnerFirst->xi2+=loserFirst->xi2;
	winnerFirst->cmx+=loserFirst->cmx;
	winnerFirst->cmy+=loserFirst->cmy;
#endif	
	return 1;

    }

    return 0;
}

